#=============================================================================
#This script runs the Bootstrap NPQIV, and studies the bootstrap SQLR 
#statistics
rm(list=ls());
#Library
require(stats);
require(MASS);
require(graphics);
require(splines);
source("Sieve_Functional_Space_Basis.R");
source("Data_Toolbox.R");
set.seed(100);
#=============================================================================
#Parameter Setting
var_sim <- matrix(c(1,0.8,0.5,0.8,1,0,0.5,0,1),3,3);
mu_sim <- c(0,0,0);
gamma <- 0.5;
alpha_smooth <- 20;
#Choose Basis
#Basis takes
#1) Pol -- Polynomial
#2) SPL -- Spline Usually cubic spline
#3) CosPol -- Cosine Polynomial
#4) SinPol -- Sine Polynomial
#5) TriPol -- Trig Polynomial
#6) HPol -- Hermite Polynomial
Basis_X <- "SPL"; Basis_Y2 <- "SPL";
output_name <- paste("NPQIV_Bootstrap", Basis_X, "_", Basis_Y2, "_1.txt");
if (Basis_X == "SPL"){
#if the conditional mean basis uses Spline, then we need to specify the order 
#of polynomial. If X_SPL_Order = r, then X basis: 1, X, X^2, ..., X^(r-1)
X_SPL_Order <- 4; #X_SPL_Order ==4 means that we have a 3rd degree polynomial.
}
#n = sample size per mc
n <- 750;
#m = number of mc studies
m <- 500;
#Boot = number of bootstrap
Boot <- 500;
#J_n = number of instruments
#Specify the J_n sequence
J_n <- 9;
#K_n = number of endogenous variables
K_n <- 6;
#lambda_n = penalization factor;
lambda_n = 0.00001;
#=============================================================================
#Penalty:
#This function calculates the C matrix of the Penalty, that is
#Pen(h) = PI'C PI
Pen_C <- function(B, B_d){
   return(t(B)%*%B+t(B_d)%*%B_d); 
} #Pen_basis
#=============================================================================
#Simulation
#=============================================================================
#Input:
#n: sample size
#m: number of iid sample
#Boot: number of bootstrap
#Ouptut:
#a n*m by Boot matrix where for each 1:n row of column b is generated by
#multinomial(n,n^-1...n^-1)
Bootstrap_Weight <- function(n,m,Boot){
 Z <- rmultinom(m*Boot, n, rep(1/n,n));
 #This returns at n by (m*Boot) matrix
    matrix_convert <- function(x){
        return(matrix(x,nrow=n));
        } #matrix_convert
    #Split the n by m*Boot matrix to n*m by Boot matrix
    return(sapply(split(Z,gl(Boot,n*m)),matrix_convert));
}
#=============================================================================
#input
#n : sample size
#m : number of iid sample
#Boot: number of boostrap replication
#var_sim : variance covariance matrix
#mu_sim  : vector of mean
#gamma   : quantile
#output
#outputs a list, with size m, where each list contains a sample of size n
DGP_MC <- function(n, m, Boot, var_sim, mu_sim, gamma){
    Z_star <- mvrnorm(n=(n*m),mu=mu_sim,Sigma=var_sim);
    Y2 <- 2*(pnorm(Z_star[,1]/3)-0.5);
    X <- 2*(pnorm(Z_star[,2]/3)-0.5);
    U <- 2*(pnorm(Z_star[,3])-gamma);
    Y1 <- 2*sin(pi*Y2)+0.76*U;
    #Z_temp is a n*m by 4 matrix
    Z_temp <- t(cbind(Y1,Y2,X,U));
    #Repeat Z matrix by Boot times.
    #and the stack the matrix
    Z_boot <- t(matrix(rep(t(Z_temp),Boot),4));
    #Now we have a n*m*Boot by 4 matrix
    ###### Weight
    #This returns a n*m By Boot matrix
    Weight <- cbind(1,Bootstrap_Weight(n,m,(Boot-1)));
    #Now stack the columns to make it a n*m*Boot by 1 matrix
    W_stack <- t(matrix(rep(Weight,1),1));
    matrix_convert <- function(x){
        return(t(matrix(x,nrow=(5))));
        } #matrix_convert
    #Stack everything together
    Z <- t(cbind(Y1,Y2,X,U,W_stack));
    return(lapply(split(Z,gl((m*Boot),(5*n))),matrix_convert));
}#end DGP_MC
Z <- DGP_MC(n,m,Boot,var_sim,mu_sim,gamma);
#=============================================================================
#PSMD optimization criteria
#input 
#PI: coefficient for unknown function
#P_proj: the projection matrix P(P'P)^(-1)P' size n by n
#Y1: n by 1 vector, dependent variable
#B : n by K_n matrix where each row is a basis of a measurable function of x
#Pen: Penalty term
#lambda: Penalization factor
#n : n, size of the sample
#alpha :smoothing parameter
#Weight vector
PSMD <- function(PI, P_proj, Y1, B, Pen, W, lambda, n, alpha = alpha_smooth){
        rho <- (Sigmoid_fun((Y1-B%*%PI), alpha)-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI)%*%Pen%*%PI)/n;
        return(Q_n);
} #end PSMD
#=============================================================================
#PSMD_Const optimization criteria with constrain
#input 
#PI: coefficient for unknown function
#P_proj: the projection matrix P(P'P)^1P' size n by n
#Y1: n by 1 vector, dependent variable
#B : n by K_n matrix where each row is a basis of a measurable function of x
#Pen: Penalty term
#lambda: Penalization factor
#n : n, size of the sample
#alpha: smoothing parameter
#basis: the unknown functional basis, since restriction will be different
#knots: knots for SPL. default to be 0.
#Y_degree: degree of polynomial, default to be 3, that is cubic spline
#r: constraint phi(h(0)) = r, default to be 0
PSMD_Const <- function(PI, P_proj, Y1, B, Pen, lambda, n, W,
    alpha = alpha_smooth, basis, knot = .4, Y_degree = 3, r=0){
        if(basis == "Pol"){
        PI_0 <- c(r, PI);
        rho <- (Sigmoid_fun((Y1-B%*%PI_0), alpha)-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI_0)%*%Pen%*%PI_0)/n;
        return(Q_n);
        } else if(basis == "SPL"){
        #The constrain is r = beta_0 + knot^Y_degree*beta_knot
        #so, beta_0 = r-knot^Y_degree*beta_knot
        N <- length(PI);
        PI_0 <- c((r-(knot^Y_degree)*PI[N-1]), PI);
        rho <- (Sigmoid_fun((Y1-B%*%PI_0), alpha)-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI_0)%*%Pen%*%PI_0)/n;
        return(Q_n);
        } else {stop("only takes Pol and SPL");}
} #end PSMD_Const
#Criterion
Q_hat_n <- function(PI, P_proj, Y1, B, Pen, lambda, n, W){
        e <- Y1-B%*%PI;
        rho <- (apply(e,2,function(x)ifelse((x<0),1,0))-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI)%*%Pen%*%PI)/n;
        return(Q_n);
}
#=============================================================================
#PSMD_MC
#This function does the Monte Carlo Simluation for the MC study 3
#input
#Z : a list of data. m iid sample of n observations
#lambda : penalization parameter
#lambda_1n : penalizatin parameter for variance
#J_n : basis order for instrumental variable X
#K_n : basis order for the unknown function
#Basis1 : type of basis approximation for instrumental variable X
#Basis2 : type of basis approximation for the unknown function
#n   : size of each sample
PSMD_MC <- function(Z, lambda, lambda_1n, J_n, K_n, Basis1, Basis2, n, r=0){
    #Data assignment
    Y1 <- Z[,1]; Y2 <- Z[,2]; X <- Z[,3]; W <- Z[,5];
    #Choosing basis for IV X
    if(Basis1 == "Pol"){
        X_r <- J_n-1;
        P <- sieve.Pol(X, X_r);
    } else if(Basis1 == "SPL"){
        X_r <- X_SPL_Order;
        knots_X <- quantile(X, (1:(J_n-X_r))/(J_n-X_r+1));
        P <- sieve.SPL(X, X_r, knots_X);
    } else if(Basis1 == "TriPol"){
        P <- sieve.TriPol(X, J_n); 
    } else if(Basis1 == "CosPol"){
        P <- sieve.CosPol(X, J_n); 
    } else if(Basis1 == "SinPol"){
        P <- sieve.SinPol(X, J_n);
    }else{stop("No Basis Function Matched");}
    #Choosing basis for unknown function
    if(Basis2 == "Pol"){
        Y2_r <- K_n-1;
        q_0_kn <- sieve.Pol(0, Y2_r);
        B <- sieve.Pol(Y2, Y2_r);
        B_d <- sieve.Pol.D(Y2, Y2_r, 1);
    } else if(Basis2 == "SPL"){
        Y2_r <- 4;
        knots_Y2 <- c(-0.4, .4);
        q_0_kn <- sieve.SPL(0, Y2_r, knots_Y2);
        B <- sieve.SPL(Y2, Y2_r, knots_Y2);
        B_d <- sieve.SPL.D(Y2, Y2_r, knots_Y2, 1);
    } else if(Basis2 == "TriPol"){
        q_0_kn <- sieve.TriPol(0, K_n);
        B <- sieve.TriPol(Y2, K_n);
        B_d <- sieve.TriPol.D(Y2,K_n,1);
    } else if(Basis2 == "CosPol"){
        q_0_kn <- sieve.CosPol(0, K_n);
        B <- sieve.CosPol(Y2, K_n);
        B_d <- sieve.CosPol.D(Y2,K_n,1);
    } else if(Basis2 == "SinPol"){
        q_0_kn <- sieve.SinPol(0, K_n);
        B <- sieve.SinPol(Y2, K_n);
        B_d <- sieve.SinPol.D(Y2,K_n,1);
    }else{stop("No Basis Function Matched");}

    ##################################################################
    #Matrix Computation
    ginv_PP <- ginv(t(P)%*%P);
    P_proj <- P%*%ginv_PP%*%t(P); 
    ##################################################################
    #Optimization
    Pen <- Pen_C(B,B_d);
    PI_in <- ginv(t(B)%*%P_proj%*%B + lambda_n*Pen_C(B,B_d))%*%
        t(B)%*%P_proj%*%Y1;
    PI_hat_temp <- optim(PI_in, PSMD, W=W, P_proj = P_proj, Y1=Y1, B=B, 
        Pen=Pen, lambda = lambda_n, n=n);
    PI_hat <- PI_hat_temp$par;
    Q_alpha_hat <- Q_hat_n(PI = PI_hat, P_proj = P_proj, Y1=Y1, B=B, n=n,
        W=W, Pen=Pen, lambda=lambda_n);
    if(Basis2 =="Pol"){
        PI0_in <- PI_in[2:K_n];
        PI_hat0_temp <- optim(PI0_in, PSMD_Const,W=W, P_proj=P_proj, Y1=Y1, 
            B=B, Pen=Pen, lambda=lambda_n, n=n,basis = Basis2);
        PI_hat0 <- PI_hat0_temp$par;
        Q_0 <- Q_hat_n(PI = c(r,PI_hat0),W=W,P_proj=P_proj,Y1=Y1,B=B, n=n,
            Pen=Pen, lambda=lambda_n);
    } else if(Basis2 =="SPL"){
        PI0_in <- PI_in[2:K_n];
        PI_hat0_temp <- optim(PI0_in, PSMD_Const,W=W,P_proj=P_proj, Y1=Y1, 
            B=B, Pen=Pen, lambda=lambda_n, n=n, basis = Basis2);
        PI_hat0 <- PI_hat0_temp$par;
        Q_0 <- Q_hat_n(PI=c((r-(knots_Y2[2]^(Y2_r-1))*PI_hat0[K_n-2]),PI_hat0),
            P_proj = P_proj, Y1=Y1, B=B, n=n, Pen=Pen,lambda=lambda_n,W=W);
    }else {stop("I only take Pol and SPL for unknown function")};
    QLR <- n*(Q_0-Q_alpha_hat);
    #Start bootstrap
    return(c(QLR, PI_hat));
} #end PSMD_MC

#=============================================================================
result_sim <- list();
model_type <- 0;
#Start the clock
ptm0 <- proc.time();
for(i in 1:length(J_n)){
    ptm <- proc.time();
    print("===========================");
    print(paste("Begin model ",i));
    result_sim[[i]] <- sapply(Z, PSMD_MC, lambda = lambda_n[i], 
    lambda_1n = lambda_1n,  J_n = J_n[i], K_n = K_n[i], Basis1 = Basis_X, 
    Basis2 = Basis_Y2, n=n);
    model_type[i] <- paste("K_n=",K_n[i]," J_n=",J_n[i], "$\\lambda_n$ =",
        lambda_n[i]);
    print("Time used: ");
    print(proc.time() - ptm);
}
#Stop the clock
proc.time() - ptm0

#=============================================================================
#rejection frequency at 1%, 5% and 10% for Chi-square
inference <- function(result,Boot,m){
    QLR_temp <- matrix(result[1,], nrow = m, ncol = Boot);
    QLR <- apply(QLR_temp,1,mean);
    r_freq_01 <- sum(QLR > qchisq(0.99,1))/(m);
    r_freq_05 <- sum(QLR > qchisq(0.95,1))/(m);
    r_freq_10 <- sum(QLR > qchisq(0.9,1))/(m);
    return(c(r_freq_01, r_freq_05, r_freq_10));
} #end Inference
#=============================================================================
#Outputs
output_name <- paste("NPQIV_Bootstrap_",Basis_X,"_",Basis_Y2,"_",
alpha_smooth, ".txt", sep="");
output_mat <- t(sapply(result_sim, inference, Boot=Boot, m=m));
colnames(output_mat) <- c("Chi_square 0.01", "0.05", "0.10");
rownames(output_mat) <- model_type;
print(output_mat);
write.table(output_mat, output_name, sep="\t");
save(result_sim, file=paste(output_name,".RData",sep=""));
#=============================================================================
#Testing only
if (m==1){
Y1 <- Z[[1]][,1];
Y2 <- Z[[1]][,2];
X <- Z[[1]][,3];
        r <- 4;
        knots_Y2 <- quantile(Y2, (1:(K_n-r))/(K_n-r+1));
    if(Basis_Y2 == "Pol"){
        B <- sieve.Pol(Y2, (K_n-1));
    } else if(Basis_Y2 == "SPL"){
        B <- sieve.SPL(Y2, r, knots_Y2);
    } else if(Basis_Y2 == "TriPol"){
        B <- sieve.TriPol(Y2, K_n);
    } else if(Basis_Y2 == "CosPol"){
        B <- sieve.CosPol(Y2, K_n);
    } else if(Basis_Y2 == "SinPol"){
        B <- sieve.SinPol(Y2, K_n);
    }else{stop("No Basis Function Matched");}
plot(Y2,Y1);
lines(sort(Y2),2*sin(pi*sort(Y2)),col=3);
PI_hat <- result_sim[[1]][2:(K_n+1)];
points(Y2,B%*%PI_hat, col = 4);
}
