#=============================================================================
#This script runs the NPIV MC studies in the Chen Pouzo Paper 2014
rm(list=ls());
#Library
require(stats);
require(MASS);
require(graphics);
require(splines);
#generate latex table
require(xtable);
source("Sieve_Functional_Space_Basis.R");
source("Data_Toolbox.R");
set.seed(100);
#sim != 0 if you want to test your model where the number of MC iteration is
#set to 1.
sim = 1;
#=============================================================================
#Parameter Setting
var_sim <- matrix(c(1,0.8,0.5,0.8,1,0,0.5,0,1),3,3);
mu_sim <- c(0,0,0);
gamma <- 0.5;
#lambda_1n in the variance derivation
lambda_1n <- 0.00001;
if (sim == 1){
#Choose Basis
#Basis takes
#1) Pol -- Polynomial
#2) SPL -- Spline Usually cubic spline
#3) CosPol -- Cosine Polynomial
#4) SinPol -- Sine Polynomial
#5) TriPol -- Trig Polynomial
#6) HPol -- Hermite Polynomial
Basis_X <- "SPL"; 
Basis_Y2 <- "SPL";
if (Basis_X == "SPL"){
#if the conditional mean basis uses Spline, then we need to specify the order 
#of polynomial. If X_SPL_Order = r, then X basis: 1, X, X^2, ..., X^(r-1)
X_SPL_Order <- 4; #X_SPL_Order ==4 means that we have a 3rd degree polynomial.
}
#n = sample size per mc
n <- 750;
#m = size of mc
m <- 5000;
#J_n = number of instruments
#Specify the J_n sequence
J_n <- c(6,9,12,15,18,21);
#K_n = number of endogenous variables 
#For cubic spline with two knots, we need K_n = 6
K_n <- rep(6,6);
#lambda_n = penalization factor;
lambda_n <- rep(0.00001,6);
}else{
#######
#model tester
n <- 750;
m <- 1;
Basis_X <- "Pol"; Basis_Y2 <- "SPL";
J_n <- 16; K_n <- 6; lambda_n = c(0.00001);
} #end else (tester)
if(length(J_n) != length(K_n)) {stop("J_n and K_n has different length");}
#=============================================================================
#Penalty:
#This function calculates the C matrix of the Penalty, that is
#Pen(h) = PI'C PI
Pen_C <- function(B, B_d){
   return(t(B)%*%B+t(B_d)%*%B_d); 
} #Pen_basis
#=============================================================================
#Simulation
#input
#n : sample size
#m : number of iid sample
#var_sim : variance covariance matrix
#mu_sim  : vector of mean
#gamma   : quantile
#output
#outputs a list, with size m, where each element of the list 
#contains a random sample of size n according to the model
#For each random sample, 1st column is Y1, follow by Y2, X and U
DGP_MC <- function(n, m, var_sim, mu_sim, gamma){
    Z_star <- mvrnorm(n=(n*m),mu=mu_sim,Sigma=var_sim);
    Y2 <- 2*(pnorm(Z_star[,1]/3)-0.5);
    X <- 2*(pnorm(Z_star[,2]/3)-0.5);
    U <- 2*(pnorm(Z_star[,3])-gamma);
    Y1 <- 2*sin(pi*Y2)+0.76*U;
    Z <- t(cbind(Y1,Y2,X,U));
    matrix_convert <- function(x){
        return(t(matrix(x,nrow=4)));
        } #matrix_convert
    return(lapply(split(Z,gl(m,4*n)),matrix_convert));
}#end DGP_MC
#Generate the list
Z <- DGP_MC(n=n,m=m,var_sim=var_sim,mu_sim=mu_sim,gamma=gamma);
#=============================================================================
#PSMD_MC
#This function does the Monte Carlo Simluation for NPIV
#input
#Z : a random sample of n observations
#lambda_n : penalization parameter
#lambda_1n : penalizatin parameter for variance
#J_n : basis order for instrumental variable X
#K_n : basis order for the unknown function
#Basis1 : type of basis approximation for instrumental variable X
#Basis2 : type of basis approximation for the unknown function
#n   : size of each sample
#Outputs
#V_1_hat (Variance 1), 
#V_2_hat (Variance 2),
#h_0_hat (\hat{h}(0)),
#Coefficient of the unknown functional PI_hat (K_n x 1) vector
PSMD_MC <- function(Z, lambda_n, lambda_1n, J_n, K_n, Basis1, Basis2, n){
    #Data assignment
    Y1 <- Z[,1]; Y2 <- Z[,2]; X <- Z[,3];
    ##########################################################################
    #Choosing basis for conditional mean functional
    if(Basis1 == "Pol"){
    #Note, I use the notation in Xiaohong's Econometrics Handbook chapter,
    #Therefore, for When J_n = 4, it means we have 3rd order polynomial,
    #that is, the basis is 1, x, x^2, x^3.
        r <- J_n-1;
        P <- sieve.Pol(X, r);
        #Note, the sieve.Pol function inputs X and r (the order of polynomial)
        #therefore, r = J_n-1
    } else if(Basis1 == "SPL"){
    #Note, again, I use the notation in Xiaohong's Econometrics Handbook,
    #Therefore,for given r, we have basis 1, x, x^2, ... x^(r-1).
    #Hence, we have (J_n-r) knots.
        r <- X_SPL_Order;
        knots_X <- quantile(X, (1:(J_n-r))/(J_n-r+1));
        P <- sieve.SPL(X, r, knots_X);
    } else if(Basis1 == "HPol"){
        r <- J_n-1;
        P <- sieve.HPol(X, r);
    } else if(Basis1 == "TriPol"){
        r <- J_n-1;
        P <- sieve.TriPol(X, r); 
    } else if(Basis1 == "CosPol"){
        r <- J_n-1;
        P <- sieve.CosPol(X, r); 
    } else if(Basis1 == "SinPol"){
        r <- J_n-1;
        P <- sieve.SinPol(X, r);
    }else{stop("No Basis Function Matched");}
    ##########################################################################
    #Choosing basis for unknown function
    if(Basis2 == "Pol"){
        r <- K_n-1;
        q_0_kn <- sieve.Pol(0, r);
        B <- sieve.Pol(Y2, r);
        B_d <- sieve.Pol.D(Y2, r, 1);
    } else if(Basis2 == "SPL"){
        r <- 4;
        #Knots are chosen at the following:
        knots_Y2 <- c(-0.4, .4);
        q_0_kn <- sieve.SPL(0, r, knots_Y2);
        B <- sieve.SPL(Y2, r, knots_Y2);
        B_d <- sieve.SPL.D(Y2, r, knots_Y2, 1);
    } else if(Basis2 == "TriPol"){
        r <- K_n-1;
        q_0_kn <- sieve.TriPol(0, r);
        B <- sieve.TriPol(Y2, r);
        B_d <- sieve.TriPol.D(Y2,r,1);
    } else if(Basis2 == "CosPol"){
        r <- K_n-1;
        q_0_kn <- sieve.CosPol(0, r);
        B <- sieve.CosPol(Y2, r);
        B_d <- sieve.CosPol.D(Y2,r,1);
    } else if(Basis2 == "SinPol"){
        r <- K_n-1;
        q_0_kn <- sieve.SinPol(0, r);
        B <- sieve.SinPol(Y2, r);
        B_d <- sieve.SinPol.D(Y2,r,1);
    }else{stop("No Basis Function Matched");}

    ##################################################################
    #Matrix Computation
    ginv_PP <- ginv(t(P)%*%P);
    P_proj <- P%*%ginv_PP%*%t(P); 
    C_n_hat <- t(B)%*%P;
    ##################################################################
    #Optimization
    #Close form solution for PI
    PI_hat <- ginv(t(B)%*%P_proj%*%B + lambda_n*Pen_C(B,B_d))%*%
        t(B)%*%P_proj%*%Y1;
    h_0_hat <- q_0_kn%*%PI_hat;
    ##################################################################
    #Variance Construction
    U_j_hat_2 <- as.vector(Y1 - B%*%PI_hat)^2;
    ###########
    #Variance 1
    D_n_hat_inverse <- solve((C_n_hat%*%ginv_PP%*%t(C_n_hat) + 
        lambda_1n*diag(K_n))/n);
    UD_Omega_n_hat <- C_n_hat%*%ginv_PP%*%t(P)%*%diag(U_j_hat_2)%*%
        P%*%ginv_PP%*%t(C_n_hat)/n;
    V_1_hat <- q_0_kn%*%D_n_hat_inverse%*%UD_Omega_n_hat%*%
        D_n_hat_inverse%*%t(q_0_kn);
    ###########
    #Variance 2
    Omega_n_hat <- C_n_hat%*%ginv_PP%*%t(P)%*%
        diag(as.vector((t(U_j_hat_2)%*%P_proj)))%*%P%*%ginv_PP%*%t(C_n_hat)/n;
    V_2_hat <- q_0_kn%*%D_n_hat_inverse%*%Omega_n_hat%*%
        D_n_hat_inverse%*%t(q_0_kn);
    return(c(V_1_hat, V_2_hat, h_0_hat, PI_hat));
} #end PSMD_MC

#=============================================================================
result_sim <- list();
model_type <- 0;
#Start the clock
ptm0 <- proc.time();
for(i in 1:length(J_n)){
    ptm <- proc.time();
    print("===========================");
    print(paste("Begin model ",i));
    result_sim[[i]] <- sapply(Z, PSMD_MC, lambda_n = lambda_n[i], 
    lambda_1n = lambda_1n,  J_n = J_n[i], K_n = K_n[i], Basis1 = Basis_X, 
    Basis2 = Basis_Y2, n=n);
    model_type[i] <- paste("K_n=",K_n[i]," J_n=",J_n[i], "$\\lambda_n$ =",
        lambda_n[i]);
    print("Time used: ");
    print(proc.time() - ptm);
}
#Stop the clock
proc.time() - ptm0

#=============================================================================
#Takes a vector V1, V2 and h_0_hat,
#outputs Med_MC described in the paper
#rejection frequency at 5% and 10%
inference <- function(result,n,m){
    V1 <- result[1,];
    V2 <- result[2,];
    h_0_hat <- result[3,];
    exp_h_0_hat <- exp(h_0_hat);
    V1_exp <- V1*exp_h_0_hat^2;
    V2_exp <- V2*exp_h_0_hat^2;
    ####################################
    #MC3
    Med_MC3_1 <- median(abs(V1/((var(h_0_hat)*n)) - 1));
    Med_MC3_2 <- median(abs(V2/((var(h_0_hat)*n)) - 1));
    t_hat_MC3_1 <- sqrt(n)*(h_0_hat)/sqrt(V1);
    t_hat_MC3_2 <- sqrt(n)*(h_0_hat)/sqrt(V2);
    r_freq_MC3_1_05 <- sum(abs(t_hat_MC3_1)>abs(qt(0.05/2,(n-1))))/m;
    r_freq_MC3_1_10 <- sum(abs(t_hat_MC3_1)>abs(qt(0.05,(n-1))))/m;
    r_freq_MC3_2_05 <- sum(abs(t_hat_MC3_2)>abs(qt(0.05/2,(n-1))))/m;
    r_freq_MC3_2_10 <- sum(abs(t_hat_MC3_2)>abs(qt(0.05,(n-1))))/m;
    ####################################
    #MC4
    Med_MC4_1 <- median(abs(V1_exp/((var(exp_h_0_hat)*n)) - 1));
    Med_MC4_2 <- median(abs(V2_exp/((var(exp_h_0_hat)*n)) - 1));
    t_hat_MC4_1 <- sqrt(n)*(exp_h_0_hat-1)/sqrt(V1_exp);
    t_hat_MC4_2 <- sqrt(n)*(exp_h_0_hat-1)/sqrt(V2_exp);
    r_freq_MC4_1_05 <- sum(abs(t_hat_MC4_1)>abs(qt(0.05/2,(n-1))))/m;
    r_freq_MC4_1_10 <- sum(abs(t_hat_MC4_1)>abs(qt(0.05,(n-1))))/m;
    r_freq_MC4_2_05 <- sum(abs(t_hat_MC4_2)>abs(qt(0.05/2,(n-1))))/m;
    r_freq_MC4_2_10 <- sum(abs(t_hat_MC4_2)>abs(qt(0.05,(n-1))))/m;
    return(c(Med_MC3_1, 
             Med_MC3_2,
             r_freq_MC3_1_05, 
             r_freq_MC3_2_05, 
             r_freq_MC3_1_10, 
             r_freq_MC3_2_10,
             Med_MC4_1, 
             Med_MC4_2,
             r_freq_MC4_1_05, 
             r_freq_MC4_2_05, 
             r_freq_MC4_1_10, 
             r_freq_MC4_2_10,
             t_hat_MC3_1,
             t_hat_MC3_2,
             t_hat_MC4_1,
             t_hat_MC4_2
             ));
} #end Inference
#=============================================================================
#Table 7.2 and 7.3
output_mat <- t(sapply(result_sim, inference, n=n, m=m));
output_mat_MC3 <- output_mat[,1:6];
output_mat_MC4 <- output_mat[,7:12];
colnames(output_mat_MC3) <- c("MC3: Med_Mc $\\hat{V}_1$",  "$\\hat{V}_2$",
    "t test 0.05 $\\hat{V}_1$", "$\\hat{V}_2$",
    "t test 0.10 $\\hat{V}_1$", "$\\hat{V}_2$");
colnames(output_mat_MC4) <- c("MC4: Med_Mc $\\hat{V}_1$",  "$\\hat{V}_2$",
    "t test 0.05 $\\hat{V}_1$", "$\\hat{V}_2$",
    "t test 0.10 $\\hat{V}_1$", "$\\hat{V}_2$");
rownames(output_mat_MC3) <- model_type;
rownames(output_mat_MC4) <- model_type;
print(output_mat_MC3);
print(output_mat_MC4);
#Output
write.table(output_mat_MC3, paste("NPIV_",Basis_X,"_",Basis_Y2,
    "_MC3.txt",sep=""),sep="\t");
write.table(output_mat_MC4, paste("NPIV_",Basis_X,"_",Basis_Y2,
    "_MC4.txt",sep=""),sep="\t");
#=============================================================================
#QQ Plot
t_v1_mc3 <- output_mat[,13];
t_v2_mc3 <- output_mat[,14];
t_v1_mc4 <- output_mat[,15];
t_v2_mc4 <- output_mat[,16];
par(mfrow=c(2,2));
qqnorm(t_v1_mc3, main="QQ Plot MC3 V1", xlim = c(-2,2), ylim = c(-2,2));
qqnorm(t_v2_mc3, main="QQ Plot MC3 V2", xlim = c(-2,2), ylim = c(-2,2));
qqnorm(t_v1_mc4, main="QQ Plot MC4 V1", xlim = c(-2,2), ylim = c(-2,2));
qqnorm(t_v2_mc4, main="QQ Plot MC4 V2", xlim = c(-2,2), ylim = c(-2,2));
#=============================================================================
#Model checking on line fit
if (m==1){
Y1 <- Z[[1]][,1];
Y2 <- Z[[1]][,2];
X <- Z[[1]][,3];
        r <- 4;
        knots_Y2 <- quantile(Y2, (1:(K_n-r))/(K_n-r+1));
    if(Basis_Y2 == "Pol"){
        B <- sieve.Pol(Y2, (K_n-1));
    } else if(Basis_Y2 == "SPL"){
        B <- sieve.SPL(Y2, r, knots_Y2);
    } else if(Basis_Y2 == "TriPol"){
        B <- sieve.TriPol(Y2, K_n);
    } else if(Basis_Y2 == "CosPol"){
        B <- sieve.CosPol(Y2, K_n);
    } else if(Basis_Y2 == "SinPol"){
        B <- sieve.SinPol(Y2, K_n);
    }else{stop("No Basis Function Matched");}
plot(Y2,Y1);
lines(sort(Y2),2*sin(pi*sort(Y2)),col=3);
PI_hat <- result_sim[[1]][4:(4+K_n-1)];
points(Y2,B%*%PI_hat, col = 4);
}
