#=============================================================================
#This script outputs the Bootstrapped SQLR power curve using Multinomial
#weight.
rm(list=ls());
#Library
require(stats);
require(MASS);
require(graphics);
require(splines);
require(xtable);
source("Sieve_Functional_Space_Basis.R");
source("Data_Toolbox.R");
seed_num = 100;
set.seed(seed_num);
#=============================================================================
#Parameter Setting
var_sim <- matrix(c(1,0.8,0.5,0.8,1,0,0.5,0,1),3,3);
mu_sim <- c(0,0,0);
gamma <- 0.5;
alpha_smooth <- 20;
#Choose Basis
#Basis takes
#1) Pol -- Polynomial
#2) SPL -- Spline Usually cubic spline
#3) CosPol -- Cosine Polynomial
#4) SinPol -- Sine Polynomial
#5) TriPol -- Trig Polynomial
#6) HPol -- Hermite Polynomial
n <- 750;
m <- 500;
Boot <- 500;
Basis_X <- "Pol"; Basis_Y2 <- "SPL";
J_n <- 10; K_n <- 6; lambda_n = 0.00001;
if (Basis_X == "SPL"){
#if the conditional mean basis uses Spline, then we need to specify the order 
#of polynomial. If X_SPL_Order = r, then X basis: 1, X, X^2, ..., X^(r-1)
X_SPL_Order <- 4; #X_SPL_Order ==4 means that we have a 3rd degree polynomial.
}
#lambda_n = penalization factor;
#Penalty:
#This function calculates the C matrix of the Penalty, that is
#Pen(h) = PI'C PI
Pen_C <- function(B, B_d){
   return(t(B)%*%B+t(B_d)%*%B_d); 
} #Pen_basis
r <- c(seq(0,0.2,0.01), 0.22,0.25,0.3);
#=============================================================================
#Simulation
#=============================================================================
#=============================================================================
#input
#n : sample size
#m : number of iid sample
#Boot: number of boostrap replication
#var_sim : variance covariance matrix
#mu_sim  : vector of mean
#gamma   : quantile
#output
#outputs a list, with size m, where each list contains a sample of size n
DGP_MC <- function(n, m, var_sim, mu_sim, gamma){
    Z_star <- mvrnorm(n=(n*m),mu=mu_sim,Sigma=var_sim);
    Y2 <- 2*(pnorm(Z_star[,1]/3)-0.5);
    X <- 2*(pnorm(Z_star[,2]/3)-0.5);
    U <- 2*(pnorm(Z_star[,3])-gamma);
    Y1 <- 2*sin(pi*Y2)+0.76*U;
    #Z_temp is a n*m by 4 matrix
    Z_temp <- t(cbind(Y1,Y2,X,U));
    #Repeat Z matrix by Boot times.
    #and the stack the matrix
    #Now we have a n*m*Boot by 4 matrix
    ###### Weight
    #This returns a n*m By Boot matrix
    matrix_convert <- function(x){
        return(t(matrix(x,nrow=(4))));
        } #matrix_convert
    #Stack everything together
    Z <- t(cbind(Y1,Y2,X,U));
    return(lapply(split(Z,gl((m),(4*n))),matrix_convert));
}#end DGP_MC

Z <- DGP_MC(n,m,var_sim,mu_sim,gamma);
#Binomial weight is assigned.
for(i in 1:m){
    W_stack <- rep(1,n);
    for(b in 1:(Boot-1)){
    W_stack <- cbind(W_stack, rmultinom(1,n,rep(1/n,n)));
    }
    Z[[i]] <- cbind(Z[[i]], W_stack);
}

#=============================================================================
#PSMD optimization criteria
#input 
#PI: coefficient for unknown function
#P_proj: the projection matrix P(P'P)^(-1)P' size n by n
#Y1: n by 1 vector, dependent variable
#B : n by K_n matrix where each row is a basis of a measurable function of x
#Pen: Penalty term
#lambda: Penalization factor
#n : n, size of the sample
#alpha :smoothing parameter
#Weight vector
PSMD <- function(PI, P_proj, Y1, B, Pen, W, lambda, n ,alpha = alpha_smooth){
        rho <- (Sigmoid_fun((Y1-B%*%PI), alpha)-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI)%*%Pen%*%PI)/n;
        return(Q_n);
} #end PSMD
#=============================================================================
#PSMD_Const optimization criteria with constrain
#input 
#PI: coefficient for unknown function
#P_proj: the projection matrix P(P'P)^1P' size n by n
#Y1: n by 1 vector, dependent variable
#B : n by K_n matrix where each row is a basis of a measurable function of x
#Pen: Penalty term
#lambda: Penalization factor
#n : n, size of the sample
#alpha: smoothing parameter
#basis: the unknown functional basis, since restriction will be different
#knots: knots for SPL. default to be 0.
#Y_degree: degree of polynomial, default to be 3, that is cubic spline
#r: constraint phi(h(0)) = r, default to be 0
PSMD_Const <- function(PI, P_proj, Y1, B, Pen, lambda, n, W,
    alpha = alpha_smooth, basis, knot = .4, Y_degree = 3, r=0){
        if(basis == "Pol"){
        PI_0 <- c(r, PI);
        rho <- (Sigmoid_fun((Y1-B%*%PI_0), alpha)-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI_0)%*%Pen%*%PI_0)/n;
        return(Q_n);
        } else if(basis == "SPL"){
        #The constrain is r = beta_0 + knot^Y_degree*beta_knot
        #so, beta_0 = r-knot^Y_degree*beta_knot
        N <- length(PI);
        PI_0 <- c((r-(knot^Y_degree)*PI[N-1]), PI);
        rho <- (Sigmoid_fun((Y1-B%*%PI_0), alpha)-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI_0)%*%Pen%*%PI_0)/n;
        return(Q_n);
        } else {stop("only takes Pol and SPL");}
} #end PSMD_Const
#Criterion
Q_hat_n <- function(PI, P_proj, Y1, B, Pen, lambda, n, W){
        e <- Y1-B%*%PI;
        rho <- (apply(e,2,function(x)ifelse((x<0),1,0))-0.5)*W;
        Q_n <- t(rho)%*%P_proj%*%(rho)/n*4
            + lambda*(t(PI)%*%Pen%*%PI)/n;
        return(Q_n);
}
#=============================================================================
#PSMD_MC
#This function does the Monte Carlo Simluation for the MC study 3
#input
#Z : a list of data. m iid sample of n observations
#lambda : penalization parameter
#J_n : basis order for instrumental variable X
#K_n : basis order for the unknown function
#Basis1 : type of basis approximation for instrumental variable X
#Basis2 : type of basis approximation for the unknown function
#n   : size of each sample
PSMD_MC <- function(Z, lambda_n, J_n, K_n, Basis1, Basis2, n, r=0){
    #Data assignment
    Y1 <- Z[,1]; Y2 <- Z[,2]; X <- Z[,3]; 
    W1 <- Z[,5]; 
    Boot_W <- Z[,6:dim(Z)[2]];
    #Choosing basis for IV X
    if(Basis1 == "Pol"){
        X_r <- J_n-1;
        P <- sieve.Pol(X, X_r);
    } else if(Basis1 == "SPL"){
        X_r <- X_SPL_Order;
        knots_X <- quantile(X, (1:(J_n-X_r))/(J_n-X_r+1));
        P <- sieve.SPL(X, X_r, knots_X);
    } else if(Basis1 == "TriPol"){
        P <- sieve.TriPol(X, J_n); 
    } else if(Basis1 == "CosPol"){
        P <- sieve.CosPol(X, J_n); 
    } else if(Basis1 == "SinPol"){
        P <- sieve.SinPol(X, J_n);
    }else{stop("No Basis Function Matched");}
    #Choosing basis for unknown function
    if(Basis2 == "Pol"){
        Y2_r <- K_n-1;
        q_0_kn <- sieve.Pol(0, Y2_r);
        B <- sieve.Pol(Y2, Y2_r);
        B_d <- sieve.Pol.D(Y2, Y2_r, 1);
    } else if(Basis2 == "SPL"){
        Y2_r <- 4;
        knots_Y2 <- c(-0.4, .4);
        q_0_kn <- sieve.SPL(0, Y2_r, knots_Y2);
        B <- sieve.SPL(Y2, Y2_r, knots_Y2);
        B_d <- sieve.SPL.D(Y2, Y2_r, knots_Y2, 1);
    } else if(Basis2 == "TriPol"){
        q_0_kn <- sieve.TriPol(0, K_n);
        B <- sieve.TriPol(Y2, K_n);
        B_d <- sieve.TriPol.D(Y2,K_n,1);
    } else if(Basis2 == "CosPol"){
        q_0_kn <- sieve.CosPol(0, K_n);
        B <- sieve.CosPol(Y2, K_n);
        B_d <- sieve.CosPol.D(Y2,K_n,1);
    } else if(Basis2 == "SinPol"){
        q_0_kn <- sieve.SinPol(0, K_n);
        B <- sieve.SinPol(Y2, K_n);
        B_d <- sieve.SinPol.D(Y2,K_n,1);
    }else{stop("No Basis Function Matched");}

    ##################################################################
    #Matrix Computation
    ginv_PP <- ginv(t(P)%*%P);
    P_proj <- P%*%ginv_PP%*%t(P); 
    ##################################################################
    #Optimization
    Pen <- Pen_C(B,B_d);
    PI_in <- ginv(t(B)%*%P_proj%*%B + lambda_n*Pen_C(B,B_d))%*%
        t(B)%*%P_proj%*%Y1;
    PI_hat_temp <- optim(PI_in, PSMD, W=W1, P_proj = P_proj, Y1=Y1, B=B, 
        Pen=Pen, lambda = lambda_n, n=n);
    PI_hat <- PI_hat_temp$par;
    Q_alpha_hat <- Q_hat_n(PI = PI_hat, P_proj = P_proj, Y1=Y1, B=B, n=n,
        W=W1, Pen=Pen, lambda=lambda_n);
    phi_hat <- q_0_kn%*%PI_hat;
    NPQIV_Const <- function(PI_in=PI_in,phi, Weight, Basis=Basis2){
    if(Basis2 =="Pol"){
        PI0_in <- PI_in[2:K_n];
        PI_hat0_temp <- optim(PI0_in, PSMD_Const,W=Weight, P_proj=P_proj, Y1=Y1, 
            r=phi, B=B, Pen=Pen, lambda=lambda_n, n=n,basis = Basis2);
        PI_hat0 <- PI_hat0_temp$par;
        Q_0 <- Q_hat_n(PI = c(phi,PI_hat0),W=W,P_proj=P_proj,Y1=Y1,B=B, n=n,
            Pen=Pen, lambda=lambda_n,W=W);
    } else if(Basis2 =="SPL"){
        PI0_in <- PI_in[2:K_n];
        PI_hat0_temp <- optim(PI0_in, PSMD_Const,W=Weight,P_proj=P_proj, Y1=Y1, 
            r=phi, B=B, Pen=Pen, lambda=lambda_n, n=n, basis = Basis2);
        PI_hat0 <- PI_hat0_temp$par;
        Q_0 <- Q_hat_n(PI=c((phi-(knots_Y2[2]^(Y2_r-1))*PI_hat0[K_n-2]),PI_hat0),
            P_proj = P_proj, Y1=Y1, B=B,n=n,Pen=Pen,lambda=lambda_n,W=Weight);
    }else {stop("I only take Pol and SPL for unknown function")};
    return(Q_0);
    } #End NPQIV_Const
    QLR <- 0;
    for(i in 1:length(r)){
    Q_0 <- NPQIV_Const(PI_in=PI_in,phi = r[i],Weight = W1);
    QLR[i] <- n*(Q_0-Q_alpha_hat);
    #Start bootstrap
    } #end for i
    for(i in 1:(Boot-1)){
    PI_hat_boot_temp <- optim(PI_in, PSMD, W=Boot_W[,i], P_proj = P_proj, 
        Y1=Y1, B=B, Pen=Pen, lambda = lambda_n, n=n);
    PI_hat_boot <- PI_hat_boot_temp$par;
    Q_B_hat <- Q_hat_n(PI = PI_hat_boot, P_proj = P_proj, Y1=Y1, B=B, n=n,
        W=Boot_W[,i], Pen=Pen, lambda=lambda_n);
    Q_0_B <- NPQIV_Const(PI_in=PI_in,phi = phi_hat,Weight = Boot_W[,i]);
    QLR[(length(r) + i)] <- n*(Q_0_B - Q_B_hat);
    }
    return(QLR);
} #end PSMD_MC
#Run
####################
#Start the clock
ptm0 <- proc.time();
    ptm <- proc.time();
    print("===========================");
    print("Begin model ");
    QLR_Mat <- sapply(Z, PSMD_MC, lambda_n = lambda_n, 
    J_n = J_n, K_n = K_n, Basis1 = Basis_X, 
    Basis2 = Basis_Y2, n=n,r=r);
    print("Time used: ");
    print(proc.time() - ptm);
#Stop the clock
proc.time() - ptm0
output_name <- paste("NPQIV_Bootstrap_", Basis_X, "_", Basis_Y2, "_", seed_num, 
".RData", sep="");
save(QLR_Mat, file = output_name);

inference <- function(QLR,mc){
    r_01 <- sum(QLR > qchisq(0.99,1))/mc;
    r_05 <- sum(QLR > qchisq(0.95,1))/mc;
    return(c(r_01, r_05));
} #end Inference
#rejection frequency at 1%, 5% and 10% for Chi-square
inference_boot_01 <- function(QLR){
    Boot_99 <- quantile(QLR[(length(r)+1):dim(QLR_Mat)[1]], 0.99);
    r_01 <- as.numeric(QLR[1:length(r)] > Boot_99); 
    return(r_01);
} #end Inference
inference_boot_05 <- function(QLR){
    Boot_95 <- quantile(QLR[(length(r)+1):dim(QLR_Mat)[1]], 0.95);
    r_05 <- as.numeric(QLR[1:length(r)] > Boot_95); 
    return(r_05);
} #end Inference

QLR_Base <- QLR_Mat[1:length(r),];
re_freq_Base <- apply(QLR_Base, 1, inference, mc = m);

re_boot_01 <- apply(QLR_Mat,2, inference_boot_01);
re_boot_05 <- apply(QLR_Mat,2, inference_boot_05);
re_freq_boot_01 <- apply(re_boot_01, 1, sum)/m;
re_freq_boot_05 <- apply(re_boot_05, 1, sum)/m;

###############################################################################
#Graphs
png(paste("Power_Curve_01_",Basis_X,"_",Basis_Y2,".png",sep=""),
    width =8, height=5,units="in",res=500);
plot(r, re_freq_boot_01 ,main="SQLR Power Curve, Rejection Probabilities
at 0.01",xlab="h(0)=r", xlim = c(0,0.3), type = "l",
    ylab="Rejection probabilities", lty=2,col="blue", yaxt = "n");
axis(2, at = seq(0,1,0.1), las=2);
lines(r, re_freq_Base[1,],lty=1,col="red");
legend("right",lty=c(2,1),c("Bootstrapped", "non-Bootstrapped"),
col=c("blue","red"));
legend("bottomright",c(paste("unknown functional: ",Basis_Y2, " K_n = 6",
sep=""),paste("conditional mean functional: ",Basis_X, " J_n = 10",sep="")));
dev.off();

png(paste("Power_Curve_05_",Basis_X,"_",Basis_Y2,".png",sep=""),
    width =8, height=5,units="in",res=500)
plot(r, re_freq_boot_05,main="SQLR Power Curve, Rejection Probabilities
at 0.05",xlab="h(0)=r", xlim = c(0,0.3),
    ylab="Rejection probabilities", type ="l", lty=2,col="blue", 
    yaxt = "n");
axis(2, at = seq(0,1,0.1), las=2);
lines(r, re_freq_Base[2,],lty=1,col="red");
legend("right",lty=c(2,1),c("Bootstrapped", "non-Bootstrapped"),
col=c("blue","red"));
legend("bottomright",c(paste("unknown functional: ",Basis_Y2, " K_n = 6",
sep=""),paste("conditional mean functional: ",Basis_X, " J_n = 10",sep="")));
dev.off();
